#!/usr/bin/env python3

from argparse import ArgumentParser, ArgumentTypeError, Action, \
    RawDescriptionHelpFormatter, BooleanOptionalAction
import re

from . import generate_job
from .store import store_data
from .action import AppendNewline, ParseKwargs, AppendKwargs


def generate_job_func(args):
    """Wrapper function for command line interface call to generate_job.

    Parameters
    ----------
        args : argparser namespace
            Command line arguments.

    Returns
    -------
        None

    """

    args_dict = {key: val for key, val in vars(args).items() if key != "func"}
    generate_job.generate_job(**args_dict)


def store_func(args, make_store=None, store_items=None, **kwargs):
    """Wrapper function for command line interface call to store.

    Parameters
    ----------
        args : argparser namespace
            Command line arguments.
        make_store : callable
            Function which generates a store object from an storm item.
        store_items : list of pairs
            List of items to store. All items need to be known to the
            make_store() function.

    Returns
    -------
        None
    """

    try:
        data_array = [[make_store(f, item, **kwargs) for item in store_items]
                      for f in args.input]
    except TypeError:
        data_array = [[make_store(item, **kwargs) for item in store_items]]

    if not args.quiet:

        print("Data:")

        for item in store_items:
            print("  - {}".format(item))

        print("{} -> {}".format('|'.join(args.input or []),
                                args.output or "stdout"))

    store_data(data_array, args.output, filters=args.filter,
               groups=args.group, append=args.append, names=args.row_names)


# Action for secondary help message
class SecondaryHelp(Action):
    def __init__(self, option_strings, dest=None, const=None, default=None,
                 help=None):
        super().__init__(option_strings=option_strings, dest=dest, const=const,
                         default=default, nargs=0, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        read_args([self.const, '--help'])


def make_parse_dict(key_type=str, val_type=str, key_only=False):
    def parse_dict(string):
        # match name and value of a equal sign separated pair
        # capture first and second string, don't capture equal sign
        if key_only:
            splt = re.match(r'(.+?)(?:(?:=)(.+))?$', string)
        else:
            splt = re.match(r'(.+?)(?:=)(.+)?$', string)

        if not splt:
            form = "'key=value'" or 'key' if key_only else 'key=value'
            raise ArgumentTypeError(f"'{string}' is not a valid variable. " +
                                    f"Expected forms like {form}")

        # extract name and value, if no value is given set to None
        key = splt.group(1)
        value = splt.group(2) or None

        return (key_type(key), None if value is None else val_type(value))
    return parse_dict


# deprecated -- use make_parse_dict
def parse_dict(string):

    # match name and value of a equal sign separated pair
    # capture first and second string, don't capture equal sign
    splt = re.match(r'(.+)(?:=)(.+)?$', string)

    if not splt:
        raise ArgumentTypeError("'" + string + "' is not a valid variable. " +
                                "Expected forms like 'key=value'")

    # extract name and value, if no value is given set to empty string
    name = splt.group(1)
    value = splt.group(2) or ""

    return (name, value)


# deprecated -- use make_parse_dict
def parse_dict_key_only(string):

    # match name and value of a equal sign separated pair
    # capture first and alternative second string, don't capture equal sign
    splt = re.match(r'(.+)((?:=)(.+))?$', string)

    if not splt:
        raise ArgumentTypeError("'" + string + "' is not a valid variable. " +
                                "Expected forms like 'key=value' or 'key'")

    # extract name and value, if no value is given set to empty string
    name = splt.group(1)
    value = splt.group(2) or ""

    return (name, value)


def parse_filter(string):
    pair = make_parse_dict(str, str)(string)
    return tuple(tuple(int(lab) for lab in item.split('/') if lab)
                 for item in pair)


def parse_occ(string):
    pair = make_parse_dict(str, str)(string)
    return (int(pair[0]), tuple(int(lab) for lab in pair[1].split('/') if lab))


def filter_parser_args(args):
    return {key: val for key, val in vars(args).items()
            if key not in ["prog", "func",  "help", "Help"]}


def parse_index(x):
    return int(x) - 1


def read_args(arg_list=None):

    description = '''
    A package for HPC applications.
    '''

    epilog = '''
    Lorem ipsum.
    '''

    parser = ArgumentParser(
        description=description,
        epilog=epilog,
        formatter_class=RawDescriptionHelpFormatter
    )

    subparsers = parser.add_subparsers()

    job = subparsers.add_parser('generate_job')

    job.set_defaults(func=generate_job_func)

    # arguments for generic job script generation; defaults allowed
    job.add_argument(
        '--profile',
        type=str,
        default='read_hostname',
        choices=["read_hostname", "csf3", "csf4", "cerberus", "medusa"],
        help="Profile specifying job submission system. \
              Default = read_hostname"
    )

    job.add_argument(
        '--submit_file',
        type=str,
        default="submit.sh",
        help='Output jobfile.'
    )

    job.add_argument(
        '--body',
        type=str,
        default="",
        action=AppendNewline,
        help='Body of the script.'
    )

    generic = job.add_argument_group('Basic arguments')

    generic.add_argument(
        '--job_name',
        type=str,
        help='Name of job.')

    generic.add_argument(
        '--env_vars',
        nargs='+',
        default={},
        type=make_parse_dict(str, str),
        action=ParseKwargs,
        help='Enviroment variables to be defined in the job script.',
        metavar='name=value')

    generic.add_argument(
        '--header',
        nargs='+',
        default=[],
        help='Custom header lines.',
        metavar='header line'
    )

    generic.add_argument(
        '--env',
        nargs='+',
        default=[],
        help='Custom environment lines.',
        metavar='env line'
    )

    generic.add_argument(
        '--exit_code',
        nargs='+',
        default=[],
        help='Custom exit code lines.',
        metavar='exit code line'
    )

    generic.add_argument(
        '--shell',
        type=str,
        default='bash',
        help='Shell to run.'
    )

    generic.add_argument(
        '--omp',
        default=1,
        type=int,
        help='Request OpenMP environment with specified number of cores.'
    )

    generic.add_argument(
        '--array',
        default=[],
        nargs='+',
        help='Create job array from list of file stems.'
    )

    generic.add_argument(
        '--verbose',
        default=True,
        action=BooleanOptionalAction,
        help='Print info message.'
    )

    # arguments for non-generic jobs; no defaults allowed
    hpc = job.add_argument_group('Submission scheduler arguments')

    hpc.add_argument(
        '--wait',
        type=bool,
        action=BooleanOptionalAction,
        help='Only return to the shell after completion of the job.'
    )

    hpc.add_argument(
        '--cwd',
        type=bool,
        action=BooleanOptionalAction,
        help='Run in current working directory.'
    )

    hpc.add_argument(
        '--pass_env',
        type=bool,
        action=BooleanOptionalAction,
        help='Pass on the env from the shell.'
    )

    hpc.add_argument(
        '--node_type',
        type=str,
        help='Type of node to request. MUST be specified on CSF3(Slurm): serial/multicore/himem.'
        'On Gadi: normal/normalbw/hugemem/hugemembw.'
    )

    hpc.add_argument(
        '--walltime',
        type=str,
        help='Walltime to request. On CSF3 (dd-hh). Default: 7-0, the maximum'
        'On Gadi: (hh:mm:ss) Default = 48:00:00, the maximum.'
    )

    hpc.add_argument(
        '--hpc_extra',
        nargs='+',
        default={},
        type=make_parse_dict(str, str),
        action=ParseKwargs,
        help='Extra flags passed to the job submission scheduler.',
        metavar='flag=value'
    )


    module = job.add_argument_group('Module arguments')

    module.add_argument(
        '--purge_modules',
        action='store_true',
        help='Run "module purge" to remove all pre-loaded modules.'
    )

    module.add_argument(
        '--modules',
        nargs='+',
        default=[],
        help='Modules to load.',
        metavar='module'
    )

    store = subparsers.add_parser('store')

    store.set_defaults(func=store_func)

    store.add_argument(
        '--quiet',
        action='store_true',
        help='disable printing of output file name to screen'
    )

    file = store.add_argument_group('file input parameters')

    file.add_argument(
        '-i', '--input',
        default=None,
        nargs='+',
        help='list of input files'
    )

    output = store.add_argument_group('output parameters')

    output.add_argument(
        '-o', '--output',
        type=str,
        help='name of the output file'
    )

    output.add_argument(
        '--group',
        type=str,
        nargs='+',
        help=('path of the root HDF5 group location where the data will be '
              'stored, one group per input')
    )

    output.add_argument(
        '--append',
        default=False,
        action='store_true',
        help='append to existing database'
    )

    output.add_argument(
        '--row_names',
        default=False,
        action='store_true',
        help='store row names'
    )

    output.add_argument(
        '--filter',
        dest='filter',
        type=parse_filter,
        default=None,
        nargs='+',
        action=AppendKwargs,
        help=('occurrences (src) of "labelled" items to keep and store at a '
              'new location (dest)'),
        metavar='src_label=dest_label'
    )

    output.add_argument(
        '--occ',
        dest='filter',
        type=parse_occ,
        default=None,
        nargs='+',
        action=AppendKwargs,
        help=('occurrences (src) of "order only" items to keep and store at a '
              'new location (dest)'),
        metavar='src_index=dest_label'
    )

    args = parser.parse_args(arg_list)

    return args


def main():
    args = read_args()
    args.func(args)



import sys
import os
import stat
import socket


class Generic:
    """
    Generic submission script class

    Attributes
    ----------
        shell : str
            Shell running the script
        verbose : bool
            Trigger to activate info message printing

    """

    def __init__(self, env_vars={}, header=[], env=[], exit_code=[],
                 job_name="", shell="bash", omp=1, verbose=True, **kwargs):
        """

        Parameters
        ----------
        env_vars : dict
            Dictionary containing name/value pairs of env variables
        header : list
            List containing the header lines
        env : list
            List containing the environment lines
        exit_code : list
            List containing the exit_code lines
        shell : str
            Shell running the script
        omp : int
            If > 1, number of cores to use in an OpenMP environment
        verbose : bool
            Trigger to activate info message printing
        """
        self.shell = shell
        self.verbose = verbose

        if self.shell == "bash":
            self.header = ["#!/bin/bash --login"] + header
            self.env = env + [
                "export {}={}".format(name, var)
                for name, var in env_vars.items()
            ]
        else:
            sys.exit("Error: Shell {} not available!".format(self.shell))

        if job_name:
            self.header += ["### Job name: {}".format(job_name)]

        self.exit_code = exit_code + ["exit $?"]

        self.unused_kwargs_warning(**kwargs)

    def unused_kwargs_warning(self, **kwargs):

        def is_empty_iterable(it):
            try:
                _ = iter(it)
                return len(it) == 0
            except TypeError:
                return False

        unused = "\n".join([
            "    {key}: {item}".format(key=key, item=item)
            for key, item in kwargs.items()
            if not (item is None or is_empty_iterable(item))
        ])

        if unused != "":
            print(
                "Warning: Unused keyword arguments in submission script ",
                "generation:\n{}".format(unused)
            )

    def write(self, f_name, body):
        """
        Function to write submission script file

        Parameters
        ----------
            f_name : str
                Name of submission script
            body : str
                Body of submission script

        Returns
        -------
            None
        """
        with open(f_name, "w") as f:

            if self.header:
                f.write("{}\n\n".format("\n".join(self.header)))

            if self.env:
                if self.shell in ["bash", "csh"]:
                    f.write("### Set up environment\n")
                f.write("{}\n\n".format("\n".join(self.env)))

            if body:
                if self.shell in ["bash", "csh"]:
                    f.write("### Main body\n")
                f.write("{}\n".format(body))

            if self.exit_code:
                f.write("{}\n\n".format("\n".join(self.exit_code)))

        # make executable
        st = os.stat(f_name)
        os.chmod(f_name, st.st_mode | stat.S_IEXEC)

    def print_message(self, f_name):
        print('The job script has been written to {}'.format(f_name))


class Modules(Generic):

    def __init__(self, purge_modules=False, modules=[], env=[], **kwargs):

        if purge_modules is None:
            _purge_modules = []
        else:
            _purge_modules = ["module purge"] if purge_modules else []

        if modules is None:
            _modules = []
        else:
            _modules = ["module load {}".format(mod) for mod in modules]

        _env = _purge_modules + _modules + env

        super().__init__(env=_env, **kwargs)

        return


class SGE(Modules):

    def __init__(self, job_name="jobname", cwd=True, pass_env=True, wait=False,
                 pe=None, hpc_extra={}, header=[], env=[], array=[], **kwargs):

        # configure sge environment
        # set variable to default value if passed None from the cli interface
        sge_conf = self.configure(
            job_name="jobname" if job_name is None else job_name,
            cwd=True if cwd is None else cwd,
            pass_env=True if pass_env is None else pass_env,
            wait=False if wait is None else wait,
            pe=None if pe is None else pe,
            array=None if array == [] else array,
            **{} if hpc_extra is None else hpc_extra
        )

        # expand header by sge configuration
        header = [
            "#$ -{} {}".format(flag, param)
            for flag, param in sge_conf.items()
        ] + header

        if array:
            env = ["TASK_ID=$SGE_TASK_ID"] + env

        if pe:
            env = ["export OMP_NUM_THREADS=$NSLOTS"] + env

        super().__init__(header=header, env=env, **kwargs)

    def configure(self, job_name="jobname", cwd=True, pass_env=True,
                  wait=False, pe=None, array=[], **hpc_extra):

        config = {'N': job_name}

        if cwd:
            config['cwd'] = ""
        if pass_env:
            config['V'] = ""
        if wait:
            config['sync'] = "yes"
        if pe is not None:
            config['pe'] = pe
        if array is not None:
            config['t'] = "1-{:d}".format(len(array))

        for key, val in hpc_extra.items():
            config[key] = val

        return config

    def print_message(self, f_name):

        print('Submit this job with:')
        print('qsub {}'.format(f_name))


class SLURM(Modules):

    def __init__(self, csf, job_name="jobname", cwd=True, pass_env=True, wait=False,
                 pe=None, hpc_extra={}, header=[], env=[], array=[], walltime=None,
                 node_type=None, **kwargs):
        
        # configure slurm environment
        # set variable to default value if passed None from the cli interface
        slurm_conf = self.configure(
            csf=csf,
            job_name="jobname" if job_name is None else job_name,
            pass_env=True if pass_env is None else pass_env,
            wait=False if wait is None else wait,
            pe=1 if pe is None else pe,
            array=None if array == [] else array,
            walltime='7-0' if walltime is None else walltime,
            partition='multicore' if node_type is None else node_type,
            **{} if hpc_extra is None else hpc_extra)

        # expand header by slurm configuration
        header = [
            "#SBATCH {}{}".format(flag, param)
            for flag, param in slurm_conf.items()
        ] + header

        if array:
            env = ["TASK_ID=$SLURM_ARRAY_TASK_ID"] + env

        if pe:
            env = ["export OMP_NUM_THREADS=$SLURM_NTASKS"] + env

        super().__init__(header=header, env=env, **kwargs)

        return

    def configure(self, csf, job_name="jobname", pass_env=True, wait=False, 
                  pe=None, partition=None, array=[], walltime='7-0', **hpc_extra):

        config = {'--job-name=': job_name}

        #set partition if given
        if csf == 4:
            if partition is not None and pe is None:
                config['-p'] = ' {}'.format(partition)
            elif partition is not None and pe is not None:
                sys.exit("Error: Cannot set partition name and OMP")
            # OMP
            if pe is not None:
                config['-p'] = ' multicore'
                config['-n'] = pe

        elif csf == 3:
            config['-t'] = ' {}'.format(walltime)
            if partition is None:
                sys.exit('Error: Node type (partition) required for CSF3')
            else:
                config['-p'] = ' {}'.format(partition) 
                config['-n'] = ' {}'.format(pe)
            
        if array is not None:
            config['--array'] = '=1-{:d}'.format(len(array))

        for key, val in hpc_extra.items():
            config[key] = val

        return config

    def print_message(self, f_name):
        
        print('Submit this job with:')
        print('sbatch {}'.format(f_name))


class PBS(Modules):

    def __init__(self, job_name="jobname", cwd=True, pass_env=True, wait=False, 
                 pe=1, mem=4, node_type=None, walltime=None, project=None, 
                 hpc_extra={}, header=[], env=[], array=[], **kwargs):

        # configure PBS environment
        # set variable to default value if passed None from the cli interface
        pbs_conf = self.configure(
            job_name="jobname" if job_name is None else job_name,
            cwd=True if cwd is None else cwd,
            pass_env=True if pass_env is None else pass_env,
            wait=False if wait is None else wait,
            pe=1 if pe is None else pe,
            mem=4 if mem is None else mem,
            node_type='normal' if node_type is None else node_type,
            # Arrays not supported on Gadi, need to include workaround for job farm
            array=None if array == [] else array,
            walltime='48:00:00' if walltime is None else walltime+':00:00',
            project='ls80' if project is None else project,
            **{} if hpc_extra is None else hpc_extra)

        # expand header by PBS configuration
        header = [
            "#PBS {}{}".format(flag, param)
            for flag, param in pbs_conf.items()
        ] + header

        # if array:
        #     env = ["TASK_ID=$SLURM_ARRAY_TASK_ID"] + env

        if pe:
            env = ["export OMP_NUM_THREADS=$PBS_NCPUS"] + env

        super().__init__(header=header, env=env, **kwargs)

        return

    def configure(self, job_name="jobname", cwd=True, pass_env=False,
                  wait=False, pe=1, mem=4, walltime='48:00:00', 
                  node_type='normal', project='ls80', array=[], **hpc_extra):

        config = {'-N ': job_name}
        # Set partition if given
        config['-q '] = '{}'.format(node_type)
        # OMP
        config['-l ncpus='] = int(pe)
        config['-l mem='] = str(mem) + 'GB'
        config['-l walltime='] = walltime
        if cwd:
            config['-l wd'] = ""
        if wait:
            config['sync'] = "yes"
        config['-P '] = project
        config['-l storage='] = 'scratch/{}+gdata/{}'.format(project,project)
        # if array is not None:
        #     config['--array'] = '=-1-{:d}'.format(len(array))

        for key, val in hpc_extra.items():
            config[key] = val
        return config

    def print_message(self, f_name):

        print('Submit this job with:')
        print('qsub {}'.format(f_name))
    
    def print_message_array(self, array_name):

        print('To submit all tasks in the array, run:')
        print('./{}'.format(array_name))

        print('To submit an individual task or list of tasks, give them as a positional argument, e.g.:')
        print('./{} 1 15 27'.format(array_name))


class oldCSF3(SGE):
    def __init__(self, omp=1, hpc_extra={}, node_type=None, **kwargs):

        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        elif omp > 1:
            pe = "smp.pe {:d}".format(omp)
        else:
            pe = None

        if node_type == "high_mem":
            hpc_extra["l"] = "mem256"
        if node_type == "short":
            hpc_extra["l"] = "short"
            if omp > 12:
                sys.exit("Cannot request >12 cores in short queue")
        super().__init__(pe=pe, hpc_extra=hpc_extra, omp=omp, **kwargs)

class CSF3(SLURM):
    def __init__(self, omp=1, hpc_extra={}, node_type=None, **kwargs):

        csf=3

        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        elif omp > 1:
            pe = " {:d}".format(omp)
        else:
            pe = None

        super().__init__(csf=csf, pe=pe,hpc_extra=hpc_extra,omp=omp,node_type=node_type,**kwargs)

class CSF4(SLURM):
    def __init__(self, omp=1, hpc_extra={}, **kwargs):

        csf=4

        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        elif omp > 1:
            pe = " {:d}".format(omp)
        else:
            pe = None

        super().__init__(csf=csf, pe=pe, hpc_extra=hpc_extra, omp=omp, **kwargs)


class Gadi(PBS):
    def __init__(self, omp=1, hpc_extra={}, node_type='normal', **kwargs):
        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        else:
            pe = " {:d}".format(omp)

        # memory, in GB
        if node_type == "hugemem":
            if omp < 6:
                sys.exit(
                    'Error: must request at least 6 cores (192GB) for {} jobs'.format(node_type))
            mem = 32*omp
        elif node_type == "hugemembw":
            if omp % 7:
                sys.exit(
                    'Error: Number of cores must be a multiple of 7 for {} jobs'.format(node_type))
            mem = int(256/7 * omp)
        elif node_type == "normalbw":
            mem = 8*omp
        else: #normal
            mem = 4*omp

        super().__init__(pe=pe, hpc_extra=hpc_extra, omp=omp,
                         mem=mem, node_type=node_type, **kwargs)


class Cerberus(Generic):
    def __init__(self, omp=1, env=[], **kwargs):

        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        elif omp > 1:
            env = ["OMP_NUM_THREADS={:d}".format(omp)] + env

        super().__init__(env=env, **kwargs)


class Medusa(Generic):
    def __init__(self, omp=1, env=[], **kwargs):

        if omp < 1:
            sys.exit('Error: Number of cores cannot be less than 1')
        elif omp > 1:
            env = ["OMP_NUM_THREADS={:d}".format(omp)] + env

        super().__init__(env=env, **kwargs)


def generate_job(profile='read_hostname', submit_file="submit.sh", body="",
                 env=[], array=[], **kwargs):
    """
    Generates jobscript for given profile (system)

    Parameters
    ----------

    profile : str, default = 'read_hostname'
        {'read_hostname', 'csf3', 'csf4', 'gadi', 'cerberus', 'medusa'}
        Name of machine
    submit_file : str, default = "submit.sh"
        Submission script filename
    body : str
        Body of submission script

    Returns
    -------
        None

    """

    gen_funcs = {
        "csf3": CSF3,
        "csf4": CSF4,
        "gadi": Gadi,
        "cerberus": Cerberus,
        "medusa": Medusa,
        }


    if profile == 'read_hostname':
        machine = parse_hostname()
    else:
        machine = profile

    if array:
        f_list = os.path.splitext(submit_file)[0] + '.txt'

        with open(f_list, "w") as f:
            for item in array:
                f.write("{}\n".format(item))
        if machine == "gadi":
            array_script_text = f'''#!/bin/bash
  
# list of tasks:
tasklist={f_list}

# jobscript name:
jobscript={submit_file}

#maximum number of tasks, from $tasklist:
num_tasks=$(wc -l < $tasklist)

#if no list of task IDs is given, submit all:
if [ -z "$1" ]
then
        echo submitting all $num_tasks tasks
        tasks=$(eval echo {{1..$num_tasks}})
else
        echo submitting tasks $@
        tasks=$@
fi

#submit each task:
for task in $tasks
do
        stem=$(sed -n "${{task}}p" "$tasklist")
        qsub -N $stem -v STEM=$stem $jobscript
done
'''
            array_script = os.path.splitext(submit_file)[0] + '.array'
            with open(array_script, "w") as f:
                f.write(array_script_text)
            # make executable
            st = os.stat(array_script)
            os.chmod(array_script, st.st_mode | stat.S_IEXEC)

        else:
            env = [r'STEM=`sed -n "${TASK_ID}p" ' + f_list + '`'] + env



    try:
        job = gen_funcs[machine](env=env, array=array, **kwargs)
    except KeyError:
        sys.exit("Error: Profile {} is not available".format(profile))

    if machine=='gadi' and array:
        job.print_message_array(array_script)
    else:
        job.print_message(submit_file)
    
    job.write(submit_file, body)

    return


def parse_hostname():
    """
    Reads hostname and detect machine type

    Parameters
    ----------
        None

    Returns
    -------
        str
            Machine type
    """

    hostname = socket.gethostname()

    if 'csf3' in hostname:
        machine = 'csf3'
    elif 'csf4' in hostname:
        machine = 'csf4'
    elif 'gadi' in hostname:
        machine = 'gadi'
    elif 'cerberus' in hostname:
        machine = 'cerberus'
    elif 'medusa' in hostname:
        machine = 'medusa'
    else:
        sys.exit("Error: Hostname unsupported, perhaps try --profile")
    return machine



from argparse import Action, _copy_items

class AppendNewline(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        # append new line if not present
        if values[-2:] != '\n':
            setattr(namespace, self.dest, values + '\n')


# todo: understand from https://sumit-ghosh.com/articles/parsing-dictionary-key-value-pairs-kwargs-argparse-python/ # noqa
class ParseKwargs(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for key, val in values:
            getattr(namespace, self.dest)[key] = val


class ParseKwargsAppend(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for key, val in values:
            if key in getattr(namespace, self.dest):
                getattr(namespace, self.dest)[key].append(val)
            else:
                getattr(namespace, self.dest)[key] = [val]


class AppendKwargs(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        items = getattr(namespace, self.dest, None)
        items = _copy_items(items)
        items.append({key: val for key, val in values})
        setattr(namespace, self.dest, items)


class OrderedSelectionAction(Action):
    def __init__(self, option_strings, dest, nargs=0, default=None, type=None,
                 choices=None, required=False, help=None, metavar=None):

        super().__init__(
            option_strings=option_strings, dest=dest, nargs=nargs,
            default=default, type=type, choices=choices, required=required,
            help=help, metavar=metavar
        )

    def __call__(self, parser, namespace, values, option_string=None):

        try:
            selection = getattr(namespace, '_selection')
        except AttributeError:
            selection = []

        selection.extend([(self.dest, val) for val in values]
                         or [(self.dest, None)])

        setattr(namespace, '_selection', selection)


# BooleanOptionalAction class until widely available in official argparse
# module. todo: remove when available
class BooleanOptionalAction(Action):
    def __init__(self, option_strings, dest, default=None, type=None,
                 choices=None, required=False, help=None, metavar=None):

        _option_strings = []
        for option_string in option_strings:
            _option_strings.append(option_string)

            if option_string.startswith('--'):
                option_string = '--no-' + option_string[2:]
                _option_strings.append(option_string)

        if help is not None and default is not None:
            help += " (default: %(default)s)"

        super().__init__(
            option_strings=_option_strings, dest=dest, nargs=0,
            default=default, type=type, choices=choices, required=required,
            help=help, metavar=metavar
        )

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest, not option_string.startswith(
                '--no-'
            ))

    def format_usage(self):
        return ' | '.join(self.option_strings)




import sys
import os
import re
from itertools import count
import mmap
import h5py
import numpy as np


def is_hdf5(f_name):
    """Predicate, which checks whether file has HDF5 extension.

    Parameters
    ----------
    f_name : str
        Name of file.

    Returns
    -------
    bool
        True if file has HDF5 extension.
    """
    if f_name:
        ext = os.path.splitext(f_name)[-1]
        return ext in ['.h5', '.hdf5']
    else:
        return False


def store_data(data_array, f_output=None, groups=None, filters=None,
               append=False, names=False):
    """Helper function to store data in output file or print to stdout.

    Parameters
    ----------
    data_array : nested list of store objects
        2D array of data sources, rows along data inputs,
        columns along data items.
    f_output : str, default=None
        Output file, default is stdout.
    filters : list of dicts, optional
        List of filter dictionaries as passed to store.write(),
        one per data item.
    groups : list of str, optional
        List of output root directories, one per data input.
    append : bool, default=False
        If True, append output to existing database.
    names : bool, default=False
        If True, print row names if available

    Returns
    -------
    None
    """

    # set up default groups = output all to root directory
    groups = [None] * len(data_array) if groups is None else groups

    if len(groups) != len(data_array):
        sys.exit("Error: Number of groups has to match number of inputs")

    if is_hdf5(f_output):
        f = h5py.File(f_output, 'r+' if append else 'w')
    elif f_output and f_output != '-':
        f = open(f_output, 'a' if append else 'w')
    else:
        f = sys.stdout

    for grp, data_list in zip(groups, data_array):

        # set up default filters = keep all
        filters = [None] * len(data_list) if filters is None else filters

        if len(filters) != len(data_list):
            sys.exit("Error: Number of filters has to match number of items")

        for fltr, data in zip(filters, data_list):
            data.write(f, fltr=fltr, root=grp, names=names)

    if f_output and f_output != '-':
        f.close()


class keep_all:
    """Filter object represented by the identity. Wraps labels of occurrence
    labelled data into tuple.

    Attributes
    ----------
    label : str or tuple
        Label of the Store object to be filtered.

    Parameters
    ----------
    label : str or tuple
        Label of the Store object to be filtered.
    """
    def __init__(self, label):
        self.label = label

    def __getitem__(self, elem):
        if self.label == "occurrence":
            return (elem,)
        else:
            return elem


class Store:
    """Base class for the storage interface for real valued data.

    Parameters
    ----------
    title : str
        Title of the dataset.
    description : str
        Description of the dataset.
    label : list of str
        Description of label items.
    units : str
        Units of data.
    fmt : str or list of str
        Global format string of data or list of format string per column.

    Attributes
    ----------
    title : str
        Title of the dataset.
    meta : dict
        Metadata, with members: description, label, units.
    fmt : str or list of str
        Global format string of data or list of format string per column.
    """
    def __init__(self, title, description, label=None, units=None, fmt=None):

        self.title = title

        self.meta = {
            'description': description,
            'label': label,
            'units': units
        }

        self.fmt = fmt

    def __iter__(self):
        """Iterator which iterates over all data items, to be overwritten by
        child class.

        Parameters
        ----------
        None

        Yields
        ------
        tuple
            Pair of label and data.
        """
        pass

    def __getitem__(self, key):
        """Access data by label.

        Parameters
        ----------
        key : tuple
            Tuple containing the label items.

        Returns
        -------
        np.array
            Numpy array containing the data.
        """
        for label, data in iter(self):
            if label == key:
                return data

        raise KeyError()

    def format_label(self, label=None, counter=None, **kwargs):
        """Standard label formatter

        Parameters
        ----------
        label : iterable
            Iterable containing label items.
        counter : itertools count object
            Count object counting the current occurrence.

        Returns
        -------
        tuple
            Tuple of label items.
        """""
        if label is None:
            if counter is None:
                return ()
            else:
                return next(counter)
        else:
            return tuple(label)

    def format_data(self, data, **kwargs):
        """Standard data formatter for real valued data.

        Parameters
        ----------
        data : array
            Data array.

        Returns
        -------
        np.array
            Numpy array containing formatted data.
        """
        return np.array(data, dtype=float)

    def filter_items(self, fltr=None):
        """Filter items based on filter dictionary.

        Parameters
        ----------
        fltr : dict, default=None
            Dictionary containing the items to be kept as keys and their
            destination as values.

        Returns
        -------
        iterator
            Iterator over filtered items.
        """
        fltr = fltr or keep_all(self.meta['label'])

        if isinstance(fltr, keep_all):
            it = ((fltr[key], val) for key, val in iter(self))
        else:
            it = ((fltr[key], val) for key, val in iter(self)
                  if key in fltr.keys())

        return it

    def write(self, file, fltr=None, root=None, names=False):
        """Write all data to HDF5 database or plain text file

        Parameters
        ----------
        file : file descriptor
            HDF5 or plain text file object pointing to the output file.
        fltr : dict, optional
            Filter dictionary containing {src: dest} entries to select only
            Specific data "src" for output to "dest".
        root : str, optional
            Root output data directory.
        names : bool, default=False
            If True, print row names if available

        Returns
        -------
        None
        """
        it = self.filter_items(fltr)

        for label, data in it:
            if isinstance(file, h5py._hl.files.File):
                self.write_h5(
                    file, label, data, root_path=root or '/', names=names)
            else:
                self.write_txt(file, label, data, names=names)

    def write_txt(self, f, label, data, names=False):
        """Default plain text dataset writer.

        Parameters
        ----------
        f : text file object
            Output file object.
        label : tuple
            Tuple of data label items.
        data : np.array
            Numpy array containing data.
        names : bool, default=False
            If True, print row names if available

        Returns
        -------
        None
        """
        header = '\n'.join(
            ["{} ({})".format(self.title, '/'.join(map(str, label)))] +
            ["{}: {}".format(key, '/'.join(val))
                if key == 'label' and val != 'occurrence' else
             "{}: {}".format(key, val) for key, val in self.meta.items()]
        )

        # write data
        if isinstance(data, dict):
            row_names, entries = zip(*data.items())
            if names:
                str_data = np.hstack(
                    (np.array(list(map(str, row_names)))[:, np.newaxis],
                     np.char.mod(self.fmt, entries)[:, np.newaxis
                     if len(np.array(entries).shape) == 1 else ...])
                )
                np.savetxt(f, str_data, fmt='%s', header=header)
            else:
                np.savetxt(f, entries, fmt=self.fmt or '%.18e', header=header)

        else:
            np.savetxt(f, data, fmt=self.fmt or '%.18e', header=header)

    def write_h5(self, h, label, data, root_path='', names=False):
        """Default HDF5 dataset writer.

        Parameters
        ----------
        h : HDF5 file object
            HDF5 output file object.
        label : tuple
            Tuple of data label items.
        data : np.array
            Numpy array containing data.
        root_path : str, optional
            Path to the HDF5 root directory.
        names : bool, default=False
            If True, print row names if available

        Returns
        -------
        None
        """
        path = self.create_h5_path(h, label, root_path)

        if isinstance(data, dict):
            row_names, data = zip(*data.items())

        if names:
            row_data = np.array(list(map(repr, row_names)), dtype='S')
            entries_data = np.array(data)
            row_meta = {"typename": type(row_names[0]).__name__,
                        "field_names": row_names[0]._fields,
                        "row_names": row_data}

            # set attribute phase change to (0, 0) in order to store all
            # attributes in dense storage to circumvent the 64kB limit
            pid = h5py.h5p.create(h5py.h5p.DATASET_CREATE)
            pid.set_attr_phase_change(0, 0)
            pid.set_attr_creation_order(h5py.h5p.CRT_ORDER_TRACKED)

            h5py.h5d.create(
                h[path].id,
                bytes(self.title, encoding="utf-8"),
                h5py.h5t.py_create(entries_data.dtype),
                h5py.h5s.create_simple(row_data.shape),
                dcpl=pid
            )

            h[path][self.title][...] = entries_data
            h[path][self.title].attrs.update(**row_meta)

        else:
            h[path].create_dataset(self.title, data=data).attrs.update(
                **self.meta)

    def create_h5_path(self, h, group, root_path=''):
        """Create HDF5 group directories and decorate with description of
        label items.

        Parameters
        ----------
        h : file object
            HDF5 file object.
        group : list
            List of group label hierarchie.
        root_path : str, optional
            Root directory of hierarchie.

        Returns
        -------
        str
            Group path.
        """
        try:
            root = h[root_path]
        except KeyError:
            h.create_group(root_path)
            root = h[root_path]

        grp_path = ""

        for idx, (mlab, lab) in enumerate(zip(self.meta['label'], group)):
            grp_path = '/'.join([str(grp) for grp in group[:(idx + 1)]])

            try:
                root.create_group(grp_path).attrs["description"] = mlab
            except ValueError:
                pass

        return '/'.join([root_path, grp_path])


class StoreType:
    """Mixin for Store class to add datatype specific members.
    """
    title: str
    meta: dict
    fmt: list

    create_h5_path: str


class Int(StoreType):
    """Storage mixin for integer valued data.
    """
    def format_data(self, data, **kwargs):
        return np.array(data, dtype=int)


class Cmplx(StoreType):
    """Storage mixin for complex valued data.
    """
    def format_data(self, real, imag, **kwargs):
        return np.array(real, dtype=float) + 1.j * np.array(imag, dtype=float)


class PlainTextExtractor(Store):
    """Plain text extractor class.

    Parameters
    ----------
    txt_file : str
        Name of extractor input.
    *args
        Arguments to be passed on to parent constructor.
    **kwargs
        Keyword arguments to be passed on to parent constructor.

    Attributes
    ----------
    txt_file : str
        Name of extractor input.
    """
    def __init__(self, txt_file, *args, **kwargs):
        self.txt_file = txt_file
        super().__init__(*args, **kwargs)

    def __iter__(self):
        # to be implemented by the child class
        pass


class RegexExtractor(PlainTextExtractor):
    """Regex extractor class.

    Parameters
    ----------
    *args
        Arguments to be passed on to parent constructor.
    **kwargs
        Keyword arguments to be passed on to parent constructor.

    Attributes
    ----------
    row_pattern : str
        Regular expression matching a single row of data. Has to contain one
        capture group per data atom.
    sec_pattern : str
        Regular expression matching the whole data section.
    """
    def __init__(self, *args, **kwargs):
        # to be implemented by child class
        self.row_pattern = None
        self.sec_pattern = None
        super().__init__(*args, **kwargs)

    def __iter__(self):

        counter = count(1)

        with open(self.txt_file, 'rb') as f:
            # use mmap to buffer file contents
            # as a result, search pattern has to be encoded to byte string
            content = mmap.mmap(f.fileno(), 0, prot=mmap.PROT_READ)
            it = re.finditer(self.sec_pattern.encode(), content)

        for m in it:
            # get dictionary of matches
            named = m.groupdict()
            # format data
            try:  # check for data array
                data = self.format_data(
                    re.findall(self.row_pattern.encode(), named["array"]),
                    **named
                )

            except KeyError:  # in case of single line of data
                data = self.format_data(None, **named)

            # generate label
            label = self.format_label(counter=counter, **named)

            yield (label, data)


class H5AttrExtractor(Store):
    """HDF5 attribute extractor class.

    Parameters
    ----------
    h_file : str
        Name of extractor input.
    h_attr : str
        Name of attribute to extract.
    h_grps : list of str
        HDF5 group paths to extract data from.
    *args
        Arguments to be passed on to parent constructor.
    **kwargs
        Keyword arguments to be passed on to parent constructor.

    Attributes
    ----------
    h_file : str
        Name of extractor input.
    h_attr : str
        Name of attribute to extract.
    h_grps : list of str
        HDF5 group paths to extract data from.
    """
    def __init__(self, h_file, h_attr, h_grps, *args, **kwargs):
        self.h_file = h_file
        self.h_attr = h_attr
        self.h_grps = h_grps
        super().__init__(*args, **kwargs)

    def __iter__(self):
        with h5py.File(self.h_file, 'r') as h:
            for grp in self.h_grps:
                label = self.format_label(re.findall(r'\/(\w+)', '/'))
                data = self.format_data(h[grp].attrs.get(self.h_attr))
                yield (label, data)


class H5Extractor(Store):
    """HDF5 dataset extractor class.

    Parameters
    ----------
    h_file : str
        Name of extractor input.
    h_dset : str
        Name of dataset to extract.
    h_grps : list of str
        HDF5 group paths to extract data from.
    *args
        Arguments to be passed on to parent constructor.
    **kwargs
        Keyword arguments to be passed on to parent constructor.

    Attributes
    ----------
    h_file : str
        Name of extractor input.
    h_dset : str
        Name of dataset to extract.
    h_grps : list of str
        HDF5 group paths to extract data from.
    """
    def __init__(self, h_file, h_dset, h_grps, *args, **kwargs):
        self.h_file = h_file
        self.h_dset = h_dset
        self.h_grps = h_grps
        super().__init__(*args, **kwargs)

    def __iter__(self):
        with h5py.File(self.h_file, 'r') as h:
            for grp in self.h_grps:
                label = self.format_label(re.findall(r'\/(\w+)', '/'))
                data = self.format_data(h[grp][self.h_dset][...])
                yield (label, data)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# pylint: disable=line-too-long,missing-module-docstring,exec-used

import setuptools

with open('README.md', 'r') as file:
    long_description = file.read()

# DO NOT EDIT THIS NUMBER!
# IT IS AUTOMATICALLY CHANGED BY python-semantic-release
__version__ = "1.12.2"

setuptools.setup(
    name='hpc-suite',
    version=__version__,
    author='Chilton Group',
    author_email='nicholas.chilton@manchester.ac.uk',
    description='A package for working with data on HPC platforms',
    url="https://gitlab.com/chilton-group/hpc_suite",
    project_urls={
        "Bug Tracker": "https://gitlab.com/chilton-group/hpc_suite/-/issues",
        "Documentation": "https://chilton-group.gitlab.io/hpc_suite"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=['numpy', 'h5py'],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'hpc_suite = hpc_suite.cli:main',
            ],
        }
)
